@@ -18,12 +18,17 @@
 #include "internal/compilers.h"
 #include "internal/hash.h"
 #include "internal/imemo.h"
+#include "internal/sanitizers.h"
 #include "node.h"
 #include "ruby/io.h"
 #include "ruby/re.h"
 #include "ruby/st.h"
 #include "symbol.h"
 
+#undef rb_funcall
+
+#include "ruby/ruby.h"
+
 /*
  *  call-seq:
  *    ObjectSpace.memsize_of(obj) -> Integer
@@ -51,32 +56,56 @@ struct total_data {
     VALUE klass;
 };
 
-static int
-total_i(void *vstart, void *vend, size_t stride, void *ptr)
+static void
+total_i(VALUE v, void *ptr)
 {
-    VALUE v;
     struct total_data *data = (struct total_data *)ptr;
 
+    if (!rb_objspace_internal_object_p(v)) {
+          if (data->klass == 0 || rb_obj_is_kind_of(v, data->klass)) {
+              data->total += rb_obj_memsize_of(v);
+          }
+    }
+}
+
+typedef void (*each_obj_with_flags)(VALUE, void*);
+
+struct obj_itr {
+    each_obj_with_flags cb;
+    void *data;
+};
+
+static int
+heap_iter(void *vstart, void *vend, size_t stride, void *ptr)
+{
+    struct obj_itr * ctx = (struct obj_itr *)ptr;
+    VALUE v;
+
     for (v = (VALUE)vstart; v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags) {
-	    switch (BUILTIN_TYPE(v)) {
-	      case T_NONE:
-	      case T_IMEMO:
-	      case T_ICLASS:
-	      case T_NODE:
-	      case T_ZOMBIE:
-		continue;
-	      default:
-		if (data->klass == 0 || rb_obj_is_kind_of(v, data->klass)) {
-		    data->total += rb_obj_memsize_of(v);
-		}
-	    }
-	}
+        void *poisoned = asan_poisoned_object_p(v);
+        asan_unpoison_object(v, false);
+
+        if (RBASIC(v)->flags) {
+            (*ctx->cb)(v, ctx->data);
+        }
+
+        if (poisoned) {
+            asan_poison_object(v);
+        }
     }
 
     return 0;
 }
 
+static void
+each_object_with_flags(each_obj_with_flags cb, void *ctx)
+{
+    struct obj_itr data;
+    data.cb = cb;
+    data.data = ctx;
+    rb_objspace_each_objects(heap_iter, &data);
+}
+
 /*
  *  call-seq:
  *    ObjectSpace.memsize_of_all([klass]) -> Integer
@@ -114,7 +143,7 @@ memsize_of_all_m(int argc, VALUE *argv, VALUE self)
 	rb_scan_args(argc, argv, "01", &data.klass);
     }
 
-    rb_objspace_each_objects(total_i, &data);
+    each_object_with_flags(total_i, &data);
     return SIZET2NUM(data.total);
 }
 
@@ -148,18 +177,11 @@ setup_hash(int argc, VALUE *argv)
     return hash;
 }
 
-static int
-cos_i(void *vstart, void *vend, size_t stride, void *data)
+static void
+cos_i(VALUE v, void *data)
 {
     size_t *counts = (size_t *)data;
-    VALUE v = (VALUE)vstart;
-
-    for (;v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags) {
-	    counts[BUILTIN_TYPE(v)] += rb_obj_memsize_of(v);
-	}
-    }
-    return 0;
+    counts[BUILTIN_TYPE(v)] += rb_obj_memsize_of(v);
 }
 
 static VALUE
@@ -236,7 +258,7 @@ count_objects_size(int argc, VALUE *argv, VALUE os)
 	counts[i] = 0;
     }
 
-    rb_objspace_each_objects(cos_i, &counts[0]);
+    each_object_with_flags(cos_i, &counts[0]);
 
     for (i = 0; i <= T_MASK; i++) {
 	if (counts[i]) {
@@ -254,25 +276,20 @@ struct dynamic_symbol_counts {
     size_t immortal;
 };
 
-static int
-cs_i(void *vstart, void *vend, size_t stride, void *n)
+static void
+cs_i(VALUE v, void *n)
 {
     struct dynamic_symbol_counts *counts = (struct dynamic_symbol_counts *)n;
-    VALUE v = (VALUE)vstart;
 
-    for (; v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_SYMBOL) {
-	    ID id = RSYMBOL(v)->id;
-	    if ((id & ~ID_SCOPE_MASK) == 0) {
-		counts->mortal++;
-	    }
-	    else {
-		counts->immortal++;
-	    }
-	}
+    if (BUILTIN_TYPE(v) == T_SYMBOL) {
+        ID id = RSYMBOL(v)->id;
+        if ((id & ~ID_SCOPE_MASK) == 0) {
+            counts->mortal++;
+        }
+        else {
+            counts->immortal++;
+        }
     }
-
-    return 0;
 }
 
 size_t rb_sym_immortal_count(void);
@@ -310,7 +327,7 @@ count_symbols(int argc, VALUE *argv, VALUE os)
     VALUE hash = setup_hash(argc, argv);
 
     size_t immortal_symbols = rb_sym_immortal_count();
-    rb_objspace_each_objects(cs_i, &dynamic_counts);
+    each_object_with_flags(cs_i, &dynamic_counts);
 
     rb_hash_aset(hash, ID2SYM(rb_intern("mortal_dynamic_symbol")),   SIZET2NUM(dynamic_counts.mortal));
     rb_hash_aset(hash, ID2SYM(rb_intern("immortal_dynamic_symbol")), SIZET2NUM(dynamic_counts.immortal));
@@ -320,20 +337,15 @@ count_symbols(int argc, VALUE *argv, VALUE os)
     return hash;
 }
 
-static int
-cn_i(void *vstart, void *vend, size_t stride, void *n)
+static void
+cn_i(VALUE v, void *n)
 {
     size_t *nodes = (size_t *)n;
-    VALUE v = (VALUE)vstart;
 
-    for (; v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_NODE) {
-	    size_t s = nd_type((NODE *)v);
-	    nodes[s]++;
-	}
+    if (BUILTIN_TYPE(v) == T_NODE) {
+        size_t s = nd_type((NODE *)v);
+        nodes[s]++;
     }
-
-    return 0;
 }
 
 /*
@@ -370,7 +382,7 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 	nodes[i] = 0;
     }
 
-    rb_objspace_each_objects(cn_i, &nodes[0]);
+    each_object_with_flags(cn_i, &nodes[0]);
 
     for (i=0; i<NODE_LAST; i++) {
 	if (nodes[i] != 0) {
@@ -404,7 +416,6 @@ count_nodes(int argc, VALUE *argv, VALUE os)
 		COUNT_NODE(NODE_MASGN);
 		COUNT_NODE(NODE_LASGN);
 		COUNT_NODE(NODE_DASGN);
-		COUNT_NODE(NODE_DASGN_CURR);
 		COUNT_NODE(NODE_GASGN);
 		COUNT_NODE(NODE_IASGN);
 		COUNT_NODE(NODE_CDECL);
@@ -493,36 +504,31 @@ count_nodes(int argc, VALUE *argv, VALUE os)
     return hash;
 }
 
-static int
-cto_i(void *vstart, void *vend, size_t stride, void *data)
+static void
+cto_i(VALUE v, void *data)
 {
     VALUE hash = (VALUE)data;
-    VALUE v = (VALUE)vstart;
 
-    for (; v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_DATA) {
-	    VALUE counter;
-	    VALUE key = RBASIC(v)->klass;
+    if (BUILTIN_TYPE(v) == T_DATA) {
+        VALUE counter;
+        VALUE key = RBASIC(v)->klass;
 
-	    if (key == 0) {
-		const char *name = rb_objspace_data_type_name(v);
-		if (name == 0) name = "unknown";
-		key = ID2SYM(rb_intern(name));
-	    }
+        if (key == 0) {
+            const char *name = rb_objspace_data_type_name(v);
+            if (name == 0) name = "unknown";
+            key = ID2SYM(rb_intern(name));
+        }
 
-	    counter = rb_hash_aref(hash, key);
-	    if (NIL_P(counter)) {
-		counter = INT2FIX(1);
-	    }
-	    else {
-		counter = INT2FIX(FIX2INT(counter) + 1);
-	    }
+        counter = rb_hash_aref(hash, key);
+        if (NIL_P(counter)) {
+            counter = INT2FIX(1);
+        }
+        else {
+            counter = INT2FIX(FIX2INT(counter) + 1);
+        }
 
-	    rb_hash_aset(hash, key, counter);
-	}
+        rb_hash_aset(hash, key, counter);
     }
-
-    return 0;
 }
 
 /*
@@ -561,37 +567,32 @@ static VALUE
 count_tdata_objects(int argc, VALUE *argv, VALUE self)
 {
     VALUE hash = setup_hash(argc, argv);
-    rb_objspace_each_objects(cto_i, (void *)hash);
+    each_object_with_flags(cto_i, (void *)hash);
     return hash;
 }
 
 static ID imemo_type_ids[IMEMO_MASK+1];
 
-static int
-count_imemo_objects_i(void *vstart, void *vend, size_t stride, void *data)
+static void
+count_imemo_objects_i(VALUE v, void *data)
 {
     VALUE hash = (VALUE)data;
-    VALUE v = (VALUE)vstart;
 
-    for (; v != (VALUE)vend; v += stride) {
-	if (RBASIC(v)->flags && BUILTIN_TYPE(v) == T_IMEMO) {
-	    VALUE counter;
-	    VALUE key = ID2SYM(imemo_type_ids[imemo_type(v)]);
+    if (BUILTIN_TYPE(v) == T_IMEMO) {
+        VALUE counter;
+        VALUE key = ID2SYM(imemo_type_ids[imemo_type(v)]);
 
-	    counter = rb_hash_aref(hash, key);
+        counter = rb_hash_aref(hash, key);
 
-	    if (NIL_P(counter)) {
-		counter = INT2FIX(1);
-	    }
-	    else {
-		counter = INT2FIX(FIX2INT(counter) + 1);
-	    }
+        if (NIL_P(counter)) {
+            counter = INT2FIX(1);
+        }
+        else {
+            counter = INT2FIX(FIX2INT(counter) + 1);
+        }
 
-	    rb_hash_aset(hash, key, counter);
-	}
+        rb_hash_aset(hash, key, counter);
     }
-
-    return 0;
 }
 
 /*
@@ -641,9 +642,10 @@ count_imemo_objects(int argc, VALUE *argv, VALUE self)
         imemo_type_ids[10] = rb_intern("imemo_parser_strterm");
         imemo_type_ids[11] = rb_intern("imemo_callinfo");
         imemo_type_ids[12] = rb_intern("imemo_callcache");
+        imemo_type_ids[13] = rb_intern("imemo_constcache");
     }
 
-    rb_objspace_each_objects(count_imemo_objects_i, (void *)hash);
+    each_object_with_flags(count_imemo_objects_i, (void *)hash);
 
     return hash;
 }
@@ -702,8 +704,8 @@ iow_internal_object_id(VALUE self)
 }
 
 struct rof_data {
-    st_table *refs;
-    VALUE internals;
+    VALUE refs;
+    VALUE values;
 };
 
 static void
@@ -714,11 +716,15 @@ reachable_object_from_i(VALUE obj, void *data_ptr)
     VALUE val = obj;
 
     if (rb_objspace_markable_object_p(obj)) {
-	if (rb_objspace_internal_object_p(obj)) {
-	    val = iow_newobj(obj);
-	    rb_ary_push(data->internals, val);
-	}
-	st_insert(data->refs, key, val);
+        if (NIL_P(rb_hash_lookup(data->refs, key))) {
+            rb_hash_aset(data->refs, key, Qtrue);
+
+            if (rb_objspace_internal_object_p(obj)) {
+                val = iow_newobj(obj);
+            }
+
+            rb_ary_push(data->values, val);
+        }
     }
 }
 
@@ -776,23 +782,21 @@ static VALUE
 reachable_objects_from(VALUE self, VALUE obj)
 {
     if (rb_objspace_markable_object_p(obj)) {
-	VALUE ret = rb_ary_new();
-	struct rof_data data;
+        struct rof_data data;
 
-	if (rb_typeddata_is_kind_of(obj, &iow_data_type)) {
-	    obj = (VALUE)DATA_PTR(obj);
-	}
+        if (rb_typeddata_is_kind_of(obj, &iow_data_type)) {
+            obj = (VALUE)DATA_PTR(obj);
+        }
 
-	data.refs = st_init_numtable();
-	data.internals = rb_ary_new();
+        data.refs = rb_obj_hide(rb_ident_hash_new());
+        data.values = rb_ary_new();
 
-	rb_objspace_reachable_objects_from(obj, reachable_object_from_i, &data);
+        rb_objspace_reachable_objects_from(obj, reachable_object_from_i, &data);
 
-	st_foreach(data.refs, collect_values, (st_data_t)ret);
-	return ret;
+        return data.values;
     }
     else {
-	return Qnil;
+        return Qnil;
     }
 }
 
@@ -895,8 +899,13 @@ objspace_internal_class_of(VALUE self, VALUE obj)
 	obj = (VALUE)DATA_PTR(obj);
     }
 
-    klass = CLASS_OF(obj);
-    return wrap_klass_iow(klass);
+    if (RB_TYPE_P(obj, T_IMEMO)) {
+        return Qnil;
+    }
+    else {
+        klass = CLASS_OF(obj);
+        return wrap_klass_iow(klass);
+    }
 }
 
 /*
@@ -983,6 +992,7 @@ Init_objspace(void)
      * You can use the #type method to check the type of the internal object.
      */
     rb_cInternalObjectWrapper = rb_define_class_under(rb_mObjSpace, "InternalObjectWrapper", rb_cObject);
+    rb_undef_alloc_func(rb_cInternalObjectWrapper);
     rb_define_method(rb_cInternalObjectWrapper, "type", iow_type, 0);
     rb_define_method(rb_cInternalObjectWrapper, "inspect", iow_inspect, 0);
     rb_define_method(rb_cInternalObjectWrapper, "internal_object_id", iow_internal_object_id, 0);
