@@ -1,44 +1,51 @@
-#include	"ruby/config.h"
+#include "ruby/config.h"
+
 #ifdef RUBY_EXTCONF_H
-#include RUBY_EXTCONF_H
+# include RUBY_EXTCONF_H
 #endif
-#include	<stdlib.h>
-#include	<stdio.h>
-#include	<sys/types.h>
-#include	<sys/stat.h>
-#include	<sys/file.h>
-#include	<fcntl.h>
-#include	<errno.h>
-#include	<pwd.h>
+
+#include <ctype.h>
+#include <errno.h>
+#include <signal.h>
+#include <stdio.h>
+#include <stdlib.h>
+
+#include <sys/types.h>
+#include <sys/stat.h>
+#include <sys/file.h>
+#include <fcntl.h>
+
+#ifdef HAVE_PWD_H
+# include <pwd.h>
+#endif
+
 #ifdef HAVE_SYS_IOCTL_H
-#include	<sys/ioctl.h>
+# include <sys/ioctl.h>
 #endif
+
 #ifdef HAVE_LIBUTIL_H
-#include	<libutil.h>
+# include <libutil.h>
 #endif
+
 #ifdef HAVE_UTIL_H
-#include	<util.h>
+# include <util.h>
 #endif
+
 #ifdef HAVE_PTY_H
-#include	<pty.h>
+# include <pty.h>
 #endif
+
 #if defined(HAVE_SYS_PARAM_H)
-  /* for __FreeBSD_version */
+ /* for __FreeBSD_version */
 # include <sys/param.h>
 #endif
+
 #ifdef HAVE_SYS_WAIT_H
-#include <sys/wait.h>
+# include <sys/wait.h>
 #else
-#define WIFSTOPPED(status)    (((status) & 0xff) == 0x7f)
+# define WIFSTOPPED(status) (((status) & 0xff) == 0x7f)
 #endif
-#include <ctype.h>
 
-#include "ruby/ruby.h"
-#include "ruby/io.h"
-#include "ruby/util.h"
-#include "internal.h"
-
-#include <signal.h>
 #ifdef HAVE_SYS_STROPTS_H
 #include <sys/stropts.h>
 #endif
@@ -47,6 +54,12 @@
 #include <unistd.h>
 #endif
 
+#include "internal.h"
+#include "internal/process.h"
+#include "internal/signal.h"
+#include "ruby/io.h"
+#include "ruby/util.h"
+
 #define	DEVICELEN	16
 
 #ifndef HAVE_SETEUID
@@ -142,9 +155,9 @@ chfunc(void *data, char *errbuf, size_t errbuf_len)
     dup2(slave,0);
     dup2(slave,1);
     dup2(slave,2);
-    close(slave);
+    if (slave < 0 || slave > 2) (void)!close(slave);
 #if defined(HAVE_SETEUID) || defined(HAVE_SETREUID) || defined(HAVE_SETRESUID)
-    seteuid(getuid());
+    if (seteuid(getuid())) ERROR_EXIT("seteuid()");
 #endif
 
     return rb_exec_async_signal_safe(carg->eargp, errbuf, sizeof(errbuf_len));
@@ -158,32 +171,32 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
     int 		master, slave, status = 0;
     rb_pid_t		pid;
     char		*p, *getenv();
-    struct passwd	*pwent;
     VALUE		v;
     struct child_info   carg;
     char		errbuf[32];
 
     if (argc == 0) {
-	const char *shellname;
+	const char *shellname = "/bin/sh";
 
 	if ((p = getenv("SHELL")) != NULL) {
 	    shellname = p;
 	}
 	else {
-	    pwent = getpwuid(getuid());
+#if defined HAVE_PWD_H
+	    const char *username = getenv("USER");
+	    struct passwd *pwent = getpwnam(username ? username : getlogin());
 	    if (pwent && pwent->pw_shell)
 		shellname = pwent->pw_shell;
-	    else
-		shellname = "/bin/sh";
+#endif
 	}
 	v = rb_str_new2(shellname);
 	argc = 1;
 	argv = &v;
     }
 
-    carg.execarg_obj = rb_execarg_new(argc, argv, 1);
+    carg.execarg_obj = rb_execarg_new(argc, argv, 1, 0);
     carg.eargp = rb_execarg_get(carg.execarg_obj);
-    rb_execarg_fixup(carg.execarg_obj);
+    rb_execarg_parent_start(carg.execarg_obj);
 
     getDevice(&master, &slave, SlaveName, 0);
 
@@ -197,12 +210,14 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
 	int e = errno;
 	close(master);
 	close(slave);
+        rb_execarg_parent_end(carg.execarg_obj);
 	errno = e;
 	if (status) rb_jump_tag(status);
 	rb_sys_fail(errbuf[0] ? errbuf : "fork failed");
     }
 
     close(slave);
+    rb_execarg_parent_end(carg.execarg_obj);
 
     info->child_pid = pid;
     info->fd = master;
@@ -210,6 +225,7 @@ establishShell(int argc, VALUE *argv, struct pty_info *info,
     RB_GC_GUARD(carg.execarg_obj);
 }
 
+#if defined(HAVE_POSIX_OPENPT) || defined(HAVE_OPENPTY) || defined(HAVE_PTSNAME)
 static int
 no_mesg(char *slavedevice, int nomesg)
 {
@@ -218,6 +234,22 @@ no_mesg(char *slavedevice, int nomesg)
     else
         return 0;
 }
+#endif
+
+#if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
+static inline int
+ioctl_I_PUSH(int fd, const char *const name)
+{
+    int ret = 0;
+# if defined(I_FIND)
+    ret = ioctl(fd, I_FIND, name);
+# endif
+    if (ret == 0) {
+        ret = ioctl(fd, I_PUSH, name);
+    }
+    return ret;
+}
+#endif
 
 static int
 get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg, int fail)
@@ -226,46 +258,38 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     /* Unix98 PTY */
     int masterfd = -1, slavefd = -1;
     char *slavedevice;
-    struct sigaction dfl, old;
 
-    dfl.sa_handler = SIG_DFL;
-    dfl.sa_flags = 0;
-    sigemptyset(&dfl.sa_mask);
-
-#if defined(__sun) || (defined(__FreeBSD__) && __FreeBSD_version < 902000)
+#if defined(__sun) || defined(__OpenBSD__) || (defined(__FreeBSD__) && __FreeBSD_version < 902000)
     /* workaround for Solaris 10: grantpt() doesn't work if FD_CLOEXEC is set.  [ruby-dev:44688] */
     /* FreeBSD 9.2 or later supports O_CLOEXEC
      * http://www.freebsd.org/cgi/query-pr.cgi?pr=162374 */
     if ((masterfd = posix_openpt(O_RDWR|O_NOCTTY)) == -1) goto error;
-    if (sigaction(SIGCHLD, &dfl, &old) == -1) goto error;
-    if (grantpt(masterfd) == -1) goto grantpt_error;
+    if (rb_grantpt(masterfd) == -1) goto error;
     rb_fd_fix_cloexec(masterfd);
 #else
     {
 	int flags = O_RDWR|O_NOCTTY;
 # if defined(O_CLOEXEC)
 	/* glibc posix_openpt() in GNU/Linux calls open("/dev/ptmx", flags) internally.
-	 * So version dependency on GNU/Linux is same as O_CLOEXEC with open().
+	 * So version dependency on GNU/Linux is the same as O_CLOEXEC with open().
 	 * O_CLOEXEC is available since Linux 2.6.23.  Linux 2.6.18 silently ignore it. */
 	flags |= O_CLOEXEC;
 # endif
 	if ((masterfd = posix_openpt(flags)) == -1) goto error;
     }
     rb_fd_fix_cloexec(masterfd);
-    if (sigaction(SIGCHLD, &dfl, &old) == -1) goto error;
-    if (grantpt(masterfd) == -1) goto grantpt_error;
+    if (rb_grantpt(masterfd) == -1) goto error;
 #endif
-    if (sigaction(SIGCHLD, &old, NULL) == -1) goto error;
     if (unlockpt(masterfd) == -1) goto error;
     if ((slavedevice = ptsname(masterfd)) == NULL) goto error;
     if (no_mesg(slavedevice, nomesg) == -1) goto error;
     if ((slavefd = rb_cloexec_open(slavedevice, O_RDWR|O_NOCTTY, 0)) == -1) goto error;
     rb_update_max_fd(slavefd);
 
-#if defined(I_PUSH) && !defined(__linux__)
-    if (ioctl(slavefd, I_PUSH, "ptem") == -1) goto error;
-    if (ioctl(slavefd, I_PUSH, "ldterm") == -1) goto error;
-    if (ioctl(slavefd, I_PUSH, "ttcompat") == -1) goto error;
+#if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
+    if (ioctl_I_PUSH(slavefd, "ptem") == -1) goto error;
+    if (ioctl_I_PUSH(slavefd, "ldterm") == -1) goto error;
+    if (ioctl_I_PUSH(slavefd, "ttcompat") == -1) goto error;
 #endif
 
     *master = masterfd;
@@ -273,8 +297,6 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
     strlcpy(SlaveName, slavedevice, DEVICELEN);
     return 0;
 
-  grantpt_error:
-    sigaction(SIGCHLD, &old, NULL);
   error:
     if (slavefd != -1) close(slavefd);
     if (masterfd != -1) close(masterfd);
@@ -326,30 +348,26 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
 
     extern char *ptsname(int);
     extern int unlockpt(int);
-    extern int grantpt(int);
 
 #if defined(__sun)
     /* workaround for Solaris 10: grantpt() doesn't work if FD_CLOEXEC is set.  [ruby-dev:44688] */
     if((masterfd = open("/dev/ptmx", O_RDWR, 0)) == -1) goto error;
-    s = signal(SIGCHLD, SIG_DFL);
-    if(grantpt(masterfd) == -1) goto error;
+    if(rb_grantpt(masterfd) == -1) goto error;
     rb_fd_fix_cloexec(masterfd);
 #else
     if((masterfd = rb_cloexec_open("/dev/ptmx", O_RDWR, 0)) == -1) goto error;
     rb_update_max_fd(masterfd);
-    s = signal(SIGCHLD, SIG_DFL);
-    if(grantpt(masterfd) == -1) goto error;
+    if(rb_grantpt(masterfd) == -1) goto error;
 #endif
-    signal(SIGCHLD, s);
     if(unlockpt(masterfd) == -1) goto error;
     if((slavedevice = ptsname(masterfd)) == NULL) goto error;
     if (no_mesg(slavedevice, nomesg) == -1) goto error;
     if((slavefd = rb_cloexec_open(slavedevice, O_RDWR, 0)) == -1) goto error;
     rb_update_max_fd(slavefd);
-#if defined(I_PUSH) && !defined(__linux__)
-    if(ioctl(slavefd, I_PUSH, "ptem") == -1) goto error;
-    if(ioctl(slavefd, I_PUSH, "ldterm") == -1) goto error;
-    ioctl(slavefd, I_PUSH, "ttcompat");
+#if defined(I_PUSH) && !defined(__linux__) && !defined(_AIX)
+    if(ioctl_I_PUSH(slavefd, "ptem") == -1) goto error;
+    if(ioctl_I_PUSH(slavefd, "ldterm") == -1) goto error;
+    ioctl_I_PUSH(slavefd, "ttcompat");
 #endif
     *master = masterfd;
     *slave = slavefd;
@@ -364,74 +382,37 @@ get_device_once(int *master, int *slave, char SlaveName[DEVICELEN], int nomesg,
 #else
     /* BSD */
     int	 masterfd = -1, slavefd = -1;
-    const char *const *p;
+    int  i;
     char MasterName[DEVICELEN];
 
-#if defined(__hpux)
-    static const char MasterDevice[] = "/dev/ptym/pty%s";
-    static const char SlaveDevice[] =  "/dev/pty/tty%s";
-    static const char *const deviceNo[] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
-    "t0","t1","t2","t3","t4","t5","t6","t7",
-    "t8","t9","ta","tb","tc","td","te","tf",
-    "u0","u1","u2","u3","u4","u5","u6","u7",
-    "u8","u9","ua","ub","uc","ud","ue","uf",
-    "v0","v1","v2","v3","v4","v5","v6","v7",
-    "v8","v9","va","vb","vc","vd","ve","vf",
-    "w0","w1","w2","w3","w4","w5","w6","w7",
-    "w8","w9","wa","wb","wc","wd","we","wf",
-    0
-    };
-#elif defined(_IBMESA)  /* AIX/ESA */
+#define HEX1(c) \
+	c"0",c"1",c"2",c"3",c"4",c"5",c"6",c"7", \
+	c"8",c"9",c"a",c"b",c"c",c"d",c"e",c"f"
+
+#if defined(_IBMESA)  /* AIX/ESA */
     static const char MasterDevice[] = "/dev/ptyp%s";
     static const char SlaveDevice[] = "/dev/ttyp%s";
-    static const char *const deviceNo[] = {
-    "00","01","02","03","04","05","06","07","08","09","0a","0b","0c","0d","0e","0f",
-    "10","11","12","13","14","15","16","17","18","19","1a","1b","1c","1d","1e","1f",
-    "20","21","22","23","24","25","26","27","28","29","2a","2b","2c","2d","2e","2f",
-    "30","31","32","33","34","35","36","37","38","39","3a","3b","3c","3d","3e","3f",
-    "40","41","42","43","44","45","46","47","48","49","4a","4b","4c","4d","4e","4f",
-    "50","51","52","53","54","55","56","57","58","59","5a","5b","5c","5d","5e","5f",
-    "60","61","62","63","64","65","66","67","68","69","6a","6b","6c","6d","6e","6f",
-    "70","71","72","73","74","75","76","77","78","79","7a","7b","7c","7d","7e","7f",
-    "80","81","82","83","84","85","86","87","88","89","8a","8b","8c","8d","8e","8f",
-    "90","91","92","93","94","95","96","97","98","99","9a","9b","9c","9d","9e","9f",
-    "a0","a1","a2","a3","a4","a5","a6","a7","a8","a9","aa","ab","ac","ad","ae","af",
-    "b0","b1","b2","b3","b4","b5","b6","b7","b8","b9","ba","bb","bc","bd","be","bf",
-    "c0","c1","c2","c3","c4","c5","c6","c7","c8","c9","ca","cb","cc","cd","ce","cf",
-    "d0","d1","d2","d3","d4","d5","d6","d7","d8","d9","da","db","dc","dd","de","df",
-    "e0","e1","e2","e3","e4","e5","e6","e7","e8","e9","ea","eb","ec","ed","ee","ef",
-    "f0","f1","f2","f3","f4","f5","f6","f7","f8","f9","fa","fb","fc","fd","fe","ff",
-    0
+    static const char deviceNo[][3] = {
+	HEX1("0"), HEX1("1"), HEX1("2"), HEX1("3"),
+	HEX1("4"), HEX1("5"), HEX1("6"), HEX1("7"),
+	HEX1("8"), HEX1("9"), HEX1("a"), HEX1("b"),
+	HEX1("c"), HEX1("d"), HEX1("e"), HEX1("f"),
     };
 #else /* 4.2BSD */
     static const char MasterDevice[] = "/dev/pty%s";
     static const char SlaveDevice[] = "/dev/tty%s";
-    static const char *const deviceNo[] = {
-    "p0","p1","p2","p3","p4","p5","p6","p7",
-    "p8","p9","pa","pb","pc","pd","pe","pf",
-    "q0","q1","q2","q3","q4","q5","q6","q7",
-    "q8","q9","qa","qb","qc","qd","qe","qf",
-    "r0","r1","r2","r3","r4","r5","r6","r7",
-    "r8","r9","ra","rb","rc","rd","re","rf",
-    "s0","s1","s2","s3","s4","s5","s6","s7",
-    "s8","s9","sa","sb","sc","sd","se","sf",
-    0
+    static const char deviceNo[][3] = {
+	HEX1("p"), HEX1("q"), HEX1("r"), HEX1("s"),
     };
 #endif
-    for (p = deviceNo; *p != NULL; p++) {
-	snprintf(MasterName, sizeof MasterName, MasterDevice, *p);
+#undef HEX1
+    for (i = 0; i < numberof(deviceNo); i++) {
+	const char *const devno = deviceNo[i];
+	snprintf(MasterName, sizeof MasterName, MasterDevice, devno);
 	if ((masterfd = rb_cloexec_open(MasterName,O_RDWR,0)) >= 0) {
             rb_update_max_fd(masterfd);
 	    *master = masterfd;
-	    snprintf(SlaveName, DEVICELEN, SlaveDevice, *p);
+	    snprintf(SlaveName, DEVICELEN, SlaveDevice, devno);
 	    if ((slavefd = rb_cloexec_open(SlaveName,O_RDWR,0)) >= 0) {
                 rb_update_max_fd(slavefd);
 		*slave = slavefd;
@@ -476,11 +457,11 @@ pty_close_pty(VALUE assoc)
 /*
  * call-seq:
  *   PTY.open => [master_io, slave_file]
- *   PTY.open {|master_io, slave_file| ... } => block value
+ *   PTY.open {|(master_io, slave_file)| ... } => block value
  *
  * Allocates a pty (pseudo-terminal).
  *
- * In the block form, yields two arguments <tt>master_io, slave_file</tt>
+ * In the block form, yields an array of two elements (<tt>master_io, slave_file</tt>)
  * and the value of the block is returned from +open+.
  *
  * The IO and File are both closed after the block completes if they haven't
@@ -504,6 +485,14 @@ pty_close_pty(VALUE assoc)
  * +slave_file+::   the slave of the pty, as a File.  The path to the
  *		    terminal device is available via +slave_file.path+
  *
+ * IO#raw! is usable to disable newline conversions:
+ *
+ *   require 'io/console'
+ *   PTY.open {|m, s|
+ *     s.raw!
+ *     # ...
+ *   }
+ *
  */
 static VALUE
 pty_open(VALUE klass)
@@ -536,8 +525,14 @@ pty_open(VALUE klass)
 }
 
 static VALUE
-pty_detach_process(struct pty_info *info)
+pty_detach_process(VALUE v)
 {
+    struct pty_info *info = (void *)v;
+#ifdef WNOHANG
+    int st;
+    if (rb_waitpid(info->child_pid, &st, WNOHANG) <= 0)
+	return Qnil;
+#endif
     rb_detach_process(info->child_pid);
     return Qnil;
 }
@@ -565,7 +560,7 @@ pty_detach_process(struct pty_info *info)
  *
  * In the block form these same values will be yielded to the block:
  *
- * +r+:: A readable IO that that contains the command's
+ * +r+:: A readable IO that contains the command's
  *       standard output and standard error
  * +w+:: A writable IO that is the command's standard input
  * +pid+:: The process identifier for the command.
@@ -585,11 +580,11 @@ pty_getpty(int argc, VALUE *argv, VALUE self)
 
     establishShell(argc, argv, &info, SlaveName);
 
-    rfptr->mode = rb_io_mode_flags("r");
+    rfptr->mode = rb_io_modestr_fmode("r");
     rfptr->fd = info.fd;
     rfptr->pathv = rb_obj_freeze(rb_str_new_cstr(SlaveName));
 
-    wfptr->mode = rb_io_mode_flags("w") | FMODE_SYNC;
+    wfptr->mode = rb_io_modestr_fmode("w") | FMODE_SYNC;
     wfptr->fd = rb_cloexec_dup(info.fd);
     if (wfptr->fd == -1)
         rb_sys_fail("dup()");
@@ -613,7 +608,7 @@ static void
 raise_from_check(rb_pid_t pid, int status)
 {
     const char *state;
-    char buf[1024];
+    VALUE msg;
     VALUE exc;
 
 #if defined(WIFSTOPPED)
@@ -631,8 +626,8 @@ raise_from_check(rb_pid_t pid, int status)
     else {
 	state = "exited";
     }
-    snprintf(buf, sizeof(buf), "pty - %s: %ld", state, (long)pid);
-    exc = rb_exc_new2(eChildExited, buf);
+    msg = rb_sprintf("pty - %s: %ld", state, (long)pid);
+    exc = rb_exc_new_str(eChildExited, msg);
     rb_iv_set(exc, "status", rb_last_status_get());
     rb_exc_raise(exc);
 }
@@ -660,15 +655,23 @@ pty_check(int argc, VALUE *argv, VALUE self)
     VALUE pid, exc;
     rb_pid_t cpid;
     int status;
+    const int flag =
+#ifdef WNOHANG
+	WNOHANG|
+#endif
+#ifdef WUNTRACED
+	WUNTRACED|
+#endif
+	0;
 
     rb_scan_args(argc, argv, "11", &pid, &exc);
-    cpid = rb_waitpid(NUM2PIDT(pid), &status, WNOHANG|WUNTRACED);
+    cpid = rb_waitpid(NUM2PIDT(pid), &status, flag);
     if (cpid == -1 || cpid == 0) return Qnil;
 
     if (!RTEST(exc)) return rb_last_status_get();
     raise_from_check(cpid, status);
 
-    UNREACHABLE;
+    UNREACHABLE_RETURN(Qnil);
 }
 
 static VALUE cPTY;
@@ -683,8 +686,8 @@ static VALUE cPTY;
 /*
  * Document-class: PTY
  *
- * Creates and managed pseudo terminals (PTYs).  See also
- * http://en.wikipedia.org/wiki/Pseudo_terminal
+ * Creates and manages pseudo terminals (PTYs).  See also
+ * https://en.wikipedia.org/wiki/Pseudo_terminal
  *
  * PTY allows you to allocate new terminals using ::open or ::spawn a new
  * terminal with a specific command.
@@ -719,7 +722,7 @@ static VALUE cPTY;
  *   # The result of read operation when pty slave is closed is platform
  *   # dependent.
  *   ret = begin
- *           m.gets          # FreeBSD returns nil.
+ *           master.gets     # FreeBSD returns nil.
  *         rescue Errno::EIO # GNU/Linux raises EIO.
  *           nil
  *         end
@@ -727,23 +730,23 @@ static VALUE cPTY;
  *
  * == License
  *
- *  C) Copyright 1998 by Akinori Ito.
+ * (c) Copyright 1998 by Akinori Ito.
  *
- *  This software may be redistributed freely for this purpose, in full
- *  or in part, provided that this entire copyright notice is included
- *  on any copies of this software and applications and derivations thereof.
+ * This software may be redistributed freely for this purpose, in full
+ * or in part, provided that this entire copyright notice is included
+ * on any copies of this software and applications and derivations thereof.
  *
- *  This software is provided on an "as is" basis, without warranty of any
- *  kind, either expressed or implied, as to any matter including, but not
- *  limited to warranty of fitness of purpose, or merchantability, or
- *  results obtained from use of this software.
+ * This software is provided on an "as is" basis, without warranty of any
+ * kind, either expressed or implied, as to any matter including, but not
+ * limited to warranty of fitness of purpose, or merchantability, or
+ * results obtained from use of this software.
  */
 
 void
-Init_pty()
+Init_pty(void)
 {
     cPTY = rb_define_module("PTY");
-    /* :nodoc */
+    /* :nodoc: */
     rb_define_module_function(cPTY,"getpty",pty_getpty,-1);
     rb_define_module_function(cPTY,"spawn",pty_getpty,-1);
     rb_define_singleton_method(cPTY,"check",pty_check,-1);
