@@ -35,7 +35,11 @@
 #    undef revents
 #  endif
 #  define RB_WAITFD_IN  POLLIN
-#  define RB_WAITFD_PRI POLLPRI
+#  if defined(POLLPRI)
+#    define RB_WAITFD_PRI POLLPRI
+#  else
+#    define RB_WAITFD_PRI 0
+#  endif
 #  define RB_WAITFD_OUT POLLOUT
 #else
 #  define RB_WAITFD_IN  0x001
@@ -670,6 +674,15 @@ VALUE rb_io_set_write_io(VALUE io, VALUE w);
  */
 void rb_io_set_nonblock(rb_io_t *fptr);
 
+/**
+ * Returns an integer representing the numeric file descriptor for
+ * <em>io</em>.
+ *
+ * @param[in]   io         An IO.
+ * @retval      int        A file descriptor.
+ */
+int rb_io_descriptor(VALUE io);
+
 /**
  * This function  breaks down the  option hash that `IO#initialize`  takes into
  * components.   This is  an implementation  detail of  rb_io_extract_modeenc()
